# $Id: wavefront.py 127 2005-11-28 21:29:25Z phf $
#
# Hacked in 2005 by Peter H. Froehlich <phf@acm.org>.
# No rights reserved.

# http://www.csit.fsu.edu/~burkardt/data/obj/obj.html
# http://www.csit.fsu.edu/~burkardt/data/mtl/mtl.html
# http://astronomy.swin.edu.au/~pbourke/dataformats/obj/
# http://www.chrisashworth.org/classes/comp236-257/final_project.php

"""
Parser for Wavefront's .obj files.

"Object files define the geometry and other properties for
objects in Wavefront's Advanced Visualizer."

The parser leaves *much* to be desired, it currently only
groks vertices and faces, nothing else. Tags the parser
doesn't understand are ignored, but you'll get warnings
about that. The code is of course quite crappy, but what
the heck, it's good enough for now. :-)

The key abstraction is the Object class which has all the
parsed information about an object. You get an instance of
Object from the load() function; eventually I should add
save() as well, but that's not a priority right now.

You can ask an Object for its vertices and faces, both of
which are lists of tuples. A vertex is a three-tuple with
the x, y, and z coordinates as given in the .obj file (so
they are floating point). A face is an n-tuple, but there
must be at least 3 vertices to talk about a face, so n>=3;
its elements are integers indexing into the vertex list.

NOTE: For some reason Wavefront likes to count from 1, but
we of course like to count from 0; so we remap all vertices
accordinly.

NOTE: For some reason certain .obj files list one and the
same vertex (in terms of coordinates) multiple times, once
for each face I guess. We prefer storing each vertex only
once, so we remap faces (or rather vertex indices) again.

TODO:

  - think through vertex and face classes; maybe there's a
    reason to do it after all? API will get heavier though

  - "negative vertex index in face" can't be parsed, was
    that premature optimization (with all the remaps and
    stuff)?

  - how should we handle object names and (named) groups?
    we currently assume that only one object is defined
    in each file...

  - clean up code... :-)
"""

# Tweak this to get debugging messages, including warnings
# about various things...
DEBUG = False

class Object:
  def __init__(self, vertices, normals, faces, name=None):
    self.normals = normals
    self.vertices = vertices
    self.faces = faces
    self.name = name

  def __str__(self):
     st = ''
     for face in self.faces:
        st = '%sFace (v, n): %s\n' % (st, face)
        for v, n in face:
           st = st + '\n      '
           st = '%s (%d)v%s  (%d)n%s, '  % (st, v, [str(round(x, 3)) for x in self.vertices[v]], n, [str(round(x, 3)) for x in self.normals[n]])
        st = st + '\n'
     return st


def load(filename, name):
  o_name = None
  vertices = []
  normals = []
  faces = []
  # line number for better (?) error and debug messages
  n_line = 0
  # number of vertices we have *seen* so far in the file
  n_seen = 0
  # number of vertices we have *accepted* so far
  n_accepted = 0
  # (x,y,z) -> integer: index we assigned for a given vertex
  accepted = {}
  # integer -> integer: given expected index, what is our index?
  mapping = {}
  # TODO
  unknowns = {}

  file = open(filename)
  n_line = 1
  for line in file:
    line = line.strip() # get rid of whitespace
    list = line.split() # split into components
    if len(list) == 0:
      if DEBUG: print "Empty: ", list
    elif list[0].startswith("#"):
      if DEBUG: print "Comment: ", list
    elif list[0] == "vn":
      #blatently copied from vertex getter
      # tricky list comprehension to grab coordinates
      normal = tuple([float(s) for s in list[1:]])
      if DEBUG: print "normal: ", normal
      # check for duplicate vertices
      if normal in accepted:
        # yep, we've already processed a normal with the same coords
        if DEBUG: print "Warning: normal repeated! ", normal
        # so we've now *seen* one more normal in the file...
        n_seen += 1
        # ...but it needs to be *mapped* to the accepted one
        mapping[n_seen] = accepted[normal]
      else:
        # nope, a genuinely "new" normal, so we'll accept it...
        accepted[normal] = n_accepted
        # ...but we still need to "map it" since we use indices
        # that are one smaller than .obj expects...
        n_seen += 1
        mapping[n_seen] = n_accepted
        # ...and we can only increment this after the mapping
        n_accepted += 1
      # finally we can append to the normal list
      normals.append(normal)
    elif list[0] == "vt":
      if DEBUG: print "TODO Vertex Texture: ", list
    elif list[0] == "v":
      # tricky list comprehension to grab coordinates
      vertex = tuple([float(s) for s in list[1:]])
      if DEBUG: print "Vertex: ", vertex
      # check for duplicate vertices
      if vertex in accepted:
        # yep, we've already processed a vertex with the same coords
        if DEBUG: print "Warning: Vertex repeated! ", vertex
        # so we've now *seen* one more vertex in the file...
        n_seen += 1
        # ...but it needs to be *mapped* to the accepted one
        mapping[n_seen] = accepted[vertex]
      else:
        # nope, a genuinely "new" vertex, so we'll accept it...
        accepted[vertex] = n_accepted
        # ...but we still need to "map it" since we use indices
        # that are one smaller than .obj expects...
        n_seen += 1
        mapping[n_seen] = n_accepted
        # ...and we can only increment this after the mapping
        n_accepted += 1
      # finally we can append to the vertex list
      vertices.append(vertex)
    elif list[0] == "f":
      face = tuple([s for s in list[1:]])
      new = []
      for v in face:
        x = v.split("/")
        if len(x) == 1:
          new.append((int(x[0])-1, -1))
        else:
          new.append((int(x[0])-1, int(x[2])-1))
      #face = [mapping[int(s)] for s in new]
      faces.append(new)
    elif list[0] == "o":
      if DEBUG: print "TODO Object name: ", list
    elif list[0] == "g":
      if DEBUG: print "TODO Group name: ", list
    else:
      if DEBUG: print "Unknown: ", list
      unknowns[list[0]] = True

    n_line += 1
  file.close()

  return Object(vertices, normals, faces, name)

if __name__ == "__main__":
  import sys as S
  model = load(S.argv[1])
  for f in model.faces():
     print f


# vim: set et sts=3 sw=3:
